/**
 * BlueTooth Caller ID  Anders Gustafsson
 *                      andersg@0x63.nu
 *                      http://0x63.nu/
 *
 * Program that connects to a Bluethooth-telephones
 * and displays status on screen using libxosd.
 *
 * 
 *  btcid is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  btcid is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with btcid; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <fcntl.h>

#include <xosd.h>

enum {
	OK=1,
	ERROR,
	CLIP,
	ELIP,
	EOLP,
	ECAV,
	CMTI,
	CKEV,
	UNKNOWN,
	EMPTY,
};

typedef union {
	struct {
		char number[64];
		int type;
	} clip;
	struct {
		char name[64];
	} elip;
	struct {
		char mem[3];
		int pos;
	} cmti;
	struct {
		int ccid;
		int ccstatus;
		int calltype;
		int processid;
		int exit_cause;
		char number[64];
		int type;
	} ecav;
	struct {
		char key[2];
		int type;
	} ckev;
	char unknown[256];
} yytype_t;

#define VERSION "0.2"

static int fd;
static yytype_t yyval; /* yes, I was thinking about using lex */

static char buf[256]; 
static int buffpos=0;

static xosd *osd;

static char *dev=NULL;
static char *command=NULL;
static int verbose=0;
static int timeout=5;
static int dodaemonize=0;
static char *fontcolor="LawnGreen";
static char *font="fixed";

static int accept(){ /* called when full line is received */
	if(buf[0]=='\0'){
		return EMPTY;
	}
	if(verbose){
		fprintf(stderr,"Received: %s\n",buf);
	}
	if(buf[0]=='+'){
		if(strncasecmp(buf,"+CLIP: ",7)==0){
			int res;
			res=sscanf(buf,"+CLIP: \"%63[^\"]\",%d",
				   yyval.clip.number,&yyval.clip.type);
			if(res==2){
				return CLIP;
			}
			res=sscanf(buf,"+CLIP: %63[^,],%d",
				   yyval.clip.number,&yyval.clip.type);
			if(res==2){
				return CLIP;
			}
			return ERROR;

		}
		if(strncasecmp(buf,"+CMTI: ",7)==0){
			int res;
			res=sscanf(buf,"+CMTI: \"%2s\",%d",
				   yyval.cmti.mem,&yyval.cmti.pos);
			if(res==2){
					return CMTI;
			}
			return ERROR;
		}
		if(strncasecmp(buf,"+CKEV: ",7)==0){
			int res;
			res=sscanf(buf,"+CKEV: %1s,%d",
				   yyval.ckev.key,&yyval.ckev.type);
			if(res==2){
					return CKEV;
			}
			return ERROR;
		}
	}
	if(buf[0]=='*'){
		if(strncasecmp(buf,"*ELIP: ",7)==0){
			int res;
			res=sscanf(buf,"*ELIP: \"%63[^\"]\"",
				   yyval.elip.name);
			if(res==1){
				return ELIP;
			}
		return ERROR;
		}
		if(strncasecmp(buf,"*EOLP: ",7)==0){
			int res;
			res=sscanf(buf,"*EOLP: \"%63[^\"]",
				   yyval.elip.name);
			if(res==1){
				return EOLP;
			}
			return ERROR;
		}
		if(strncasecmp(buf,"*ECAV: ",7)==0){
			int res;
			res=sscanf(buf,"*ECAV: %d,%d,%d,%d,%d,\"%63[^\"],%d",
				   &yyval.ecav.ccid,
				   &yyval.ecav.ccstatus,
				   &yyval.ecav.calltype,
				   &yyval.ecav.processid,
				   &yyval.ecav.exit_cause,
				   yyval.ecav.number,
				   &yyval.ecav.type);
			if(res==3 && yyval.ecav.ccstatus>1){
				return ECAV;
			}
			if(res==5 && yyval.ecav.ccstatus==0){
				return ECAV;
			}
			res=sscanf(buf,"*ECAV: %d,%d,%d,,,\"%63[^\"]\",%d",
				   &yyval.ecav.ccid,
				   &yyval.ecav.ccstatus,
				   &yyval.ecav.calltype,
				   yyval.ecav.number,
				   &yyval.ecav.type);
			if(res==5){
				return ECAV;
			}

			return ERROR;
		}

	}
	if(strcasecmp(buf,"OK")==0){
		return OK;
	}
	if(strcasecmp(buf,"ERROR")==0){
		return ERROR;
	}
	strcpy(yyval.unknown,buf);
	return UNKNOWN;
}

static int yylex(){
	int i;

	while(57){
		for(i=0;i<buffpos;i++){
			if(buf[i]=='\n' || buf[i]=='\r'){
				int type;
				buf[i]='\0';
				type=accept();
				memmove(buf,buf+i+1,buffpos-i-1);
				buffpos-=i+1;
				return type;
			}
		}
		i=read(fd,buf+buffpos,127);
		if(i>0){
			buffpos+=i;
		}
	}
/* 	return ERROR;   never reached */
}

static void execute(char *cmd){
	write(fd,"\r\n",2);
	write(fd,cmd,strlen(cmd));
	write(fd,"\r\n",2);

	if(verbose){
		fprintf(stderr,"Sending: %s\n",cmd);
	}

	// res=yylex();  /* take care of result */
	sleep(1);
}

static void setup(){
	struct termios opts;
	fd=open(dev,O_RDWR);
	if(fd==-1){
		perror("Couldn't open device");
		exit(EXIT_FAILURE);
	}

	/* setup device */
	tcgetattr(fd,&opts);
	opts.c_lflag &= ~(ICANON | ECHO | ISIG);
	opts.c_cflag |= (CLOCAL | CREAD);
	cfsetispeed(&opts,B38400);
	opts.c_iflag |= (INPCK | ISTRIP);
	opts.c_iflag |= (IXON | IXOFF | IXANY);
	opts.c_cflag &= ~CSIZE;
	opts.c_cflag |= CS8;
	opts.c_oflag &= ~OPOST;
	opts.c_cc[VMIN] = 1;
	opts.c_cc[VTIME] = 0;
	tcsetattr(fd,TCSANOW,&opts);

	/* init device */
	sleep(1);
	execute("AT&F");        /* factory configuration */
	execute("ATE0");        /* no echoing */
	execute("AT+CSCS=\"8859-1\""); /* charset */
	execute("AT+CLIP=1");   /* enable callid notifications */
	execute("AT+CNMI=3,1"); /* enable sms notifications */
	execute("AT*EIPS=1,1"); /* enable name callid */
	execute("AT*EIPS=2,1"); /* enable name callout id */
	execute("AT*ECAM=1");   /* enable call monitoring */
	// useful for debugging
	//execute("AT+CMER=3,2"); /* key reporting */

}

static void run(){
	if(command!=NULL){
		if(verbose){
			fprintf(stderr,"Executing: %s\n",command);
		}
		system(command); /* change to fork/exec so its run async? */
	}
}

static void show_usage(){
	puts("btcid recognizes the following options:");
	puts(" -p port  --port port             device to attach to");
	puts(" -c cmd   --command cmd           command to execute on call");
	puts(" -V       --version               display version and exit");
	puts(" -v       --verbose               be verbose");
	puts(" -f       --font                  specify font");
	puts(" -a       --fontcolor             set fontcolor");
	puts(" -d       --daemonize             fork to background");
	puts(" -t       --timeout               set OSD timeout\n");
}

static void parse_options(int argc,char **argv){
	static struct option long_options[] = {
		{"port", 1, 0, 'p'},
		{"command", 1, 0, 'c'},
		{"version", 0, 0, 'V'},
		{"verbose", 0, 0, 'v'},
		{"help", 0, 0, 'h'},
		{"fontcolor", 1, 0, 'a'},
		{"font", 1, 0, 'f'},
		{"timeout", 1, 0, 't'},
		{"daemonize", 0, 0, 'd'},
		{0, 0, 0, 0}
	};
	int c;

	do{
		c=getopt_long(argc,argv,"p:c:t:f:a:dVv",long_options,NULL);
		switch(c){
		case 'p':
			dev=optarg;
			break;
		case 'c':
			command=optarg;
			break;
		case 'V':
			printf("btcid " VERSION "\n");
			exit(EXIT_SUCCESS);
			break;
		case 'v':
			verbose=1;
			break;
		case 't':
			timeout=atoi(optarg);
			break;
		case 'f':
			font=optarg;
			break;
		case 'a':
			fontcolor=optarg;
			break;
		case 'd':
			dodaemonize=1;
			break;
		case 'h':
			show_usage();
			exit(EXIT_SUCCESS);
			break;
		}
	}while(c!=-1);

	if(optind!=argc){
		show_usage();
		exit(EXIT_FAILURE);
	}
}

int main(int argc,char **argv){
	int tok;
	char tmp[128];
	
	parse_options(argc,argv);

	if(dev==NULL){
		fprintf(stderr,"No port specified\n");
		exit(EXIT_FAILURE);
	}

	if(dodaemonize){
		pid_t pid;

		close(STDIN_FILENO);
		close(STDOUT_FILENO);
		close(STDERR_FILENO);

		freopen("/dev/null", "r", stdin);
		freopen("/dev/null", "w", stdout);
		freopen("/dev/null", "w", stderr);
	   
		pid=fork();
		if(pid){
			exit (0);
		}
	}

	setup();

	osd=xosd_init(font, fontcolor, timeout, XOSD_bottom, 5, 1, 2);

	if(osd==NULL){
		if(!dodaemonize){
			fprintf(stderr,"couldn't init xosd\n");
		}
		exit(EXIT_FAILURE);
	}

	xosd_display(osd, 0, XOSD_string, "Bluetooth CallerID version " VERSION " started!");

	/* main loop */
	while((tok=yylex())){
		switch(tok){
		case CLIP:{
			snprintf(tmp,127,"Incoming call from: %s",yyval.clip.number);
			xosd_display(osd, 0, XOSD_string, tmp);
			run();
			break;
		}
		case ELIP:{
			xosd_display(osd, 1, XOSD_string, yyval.elip.name);
			break;
		}
		case EOLP:{
			xosd_display(osd, 1, XOSD_string, yyval.elip.name);
			break;
		}
		case CMTI:{
			int tmppos;
			tmppos=yyval.cmti.pos;
			snprintf(tmp,127,"New sms, in '%s' pos: %d",yyval.cmti.mem,tmppos);
			xosd_display(osd, 0, XOSD_string, tmp);
			xosd_display(osd, 1, XOSD_string, "");
			snprintf(tmp,127,"AT+CPMS=\"%2s\"",yyval.cmti.mem);
			execute(tmp); /* might destroy yyval */
			snprintf(tmp,127,"AT+CMGR=%d",tmppos);
			execute(tmp);
			break;
		}
		case CKEV:{
			printf("key %s:%s\n",yyval.ckev.type==1?"up":"down",yyval.ckev.key);
			break;
		}

		case ECAV:{
			if(yyval.ecav.ccstatus==0){
				xosd_display(osd, 0, XOSD_string, "Call ended");
				xosd_display(osd, 1, XOSD_string, "");
			}
			if(yyval.ecav.ccstatus==1){
				snprintf(tmp,127,"Calling %s...",yyval.ecav.number);
				xosd_display(osd, 0, XOSD_string, tmp);
				run();
			}
			if(yyval.ecav.ccstatus==2){
				xosd_display(osd, 0, XOSD_string, "Connecting...");
			}
			if(yyval.ecav.ccstatus==3){
				xosd_display(osd, 0, XOSD_string, "Connected!");
			}
			if(yyval.ecav.ccstatus==7){
				xosd_display(osd, 0, XOSD_string, "Busy!");
			}
		}
			
		}
	}
	
	xosd_uninit(osd);
	
	return 0;
}
